from collections import OrderedDict
import hug.types


def generate(module, base_url=""):
    documentation = OrderedDict()
    documentation['overview'] = module.__doc__
    for url, method_handler in module.HUG_API_CALLS.items():
        url_doc = documentation.setdefault(url, {})

        mapping = OrderedDict()
        for method, handler in method_handler.items():
            mapping.setdefault(handler, []).append(method.split("_")[-1].upper())

        for handler, methods in mapping.items():
            doc = url_doc.setdefault(",".join(methods), OrderedDict())
            doc['usage'] = handler.api_function.__doc__
            if handler.example:
                doc['example'] = "{0}{1}".format(base_url, url)
                if isinstance(handler.example, str):
                    doc['example'] += "?{0}".format(handler.example)
            inputs = doc.setdefault('inputs', OrderedDict())
            doc['outputs'] = OrderedDict(format=handler.output_format.__doc__)

            api = handler.api_function
            types = api.__annotations__
            arguments = api.__code__.co_varnames[:api.__code__.co_argcount]

            defaults = {}
            for index, default in enumerate(api.__defaults__ or ()):
                defaults[arguments[-(index + 1)]] = default

            for argument in arguments:
                if argument in ('request', 'response'):
                    continue

                input_definition = inputs.setdefault(argument, OrderedDict())
                input_definition['type'] = types.get(argument, hug.types.text).__doc__
                default = defaults.get(argument, None)
                if default is not None:
                    input_definition['default'] = default

    return documentation




import json as json_converter


def json(content):
    """JSON (Javascript Serialized Object Notation)"""
    return json_converter.dumps(content).encode('utf8')


def text(content):
    """Free form UTF8 text"""
    return content.encode('utf8')


def _camelcase(dictionary):
    if not isinstance(dictionary, dict):
        return dictionary

    new_dictionary = {}
    for key, value in dictionary.items():
        if isinstance(key, str):
            key = key[0] + "".join(key.title().split('_'))[1:]
        new_dictionary[key] = _camelcase(dictionary)


def json_camelcase(content):
    """JSON (Javascript Serialized Object Notation) with all keys camelCased"""
    return _camelcase(json(body))



import json as json_converter
import re

UNDERSCORE = (re.compile('(.)([A-Z][a-z]+)'), re.compile('([a-z0-9])([A-Z])'))


def json(body):
    return json_converter.loads(body)


def _under_score_dict(dictionary):
    new_dictionary = {}
    for key, value in dictionary.items():
        if isinstance(value, dict):
            value = _under_score_dict(dictionary)
        if isinstance(key, str):
            key = UNDERSCORE[1].sub(r'\1_\2', UNDERSCORE[0].sub(r'\1_\2', name)).lower()
        new_dictionary[key] = value


def json_underscore(body):
    return _under_score_dict(json(body))



from functools import wraps, partial
from collections import OrderedDict
import sys
from hug.run import server
import hug.output_format

from falcon import HTTP_METHODS, HTTP_BAD_REQUEST


def call(urls, accept=HTTP_METHODS, output=hug.output_format.json, example=None):
    if isinstance(urls, str):
        urls = (urls, )

    def decorator(api_function):
        module = sys.modules[api_function.__module__]

        def interface(request, response):
            input_parameters = request.params
            errors = {}
            for key, type_handler in api_function.__annotations__.items():
                try:
                    input_parameters[key] = type_handler(input_parameters[key])
                except Exception as error:
                    errors[key] = str(error)
            if errors:
                response.data = output({"errors": errors})
                response.status = HTTP_BAD_REQUEST
                return

            input_parameters['request'], input_parameters['response'] = (request, response)
            response.data = output(api_function(**input_parameters))

        if not 'HUG' in module.__dict__:
            def api_auto_instantiate(*kargs, **kwargs):
                module.HUG = server(module)
                return module.HUG(*kargs, **kwargs)
            module.HUG = api_auto_instantiate
            module.HUG_API_CALLS = OrderedDict()

        for url in urls:
            handlers = module.HUG_API_CALLS.setdefault(url, {})
            for method in accept:
                handlers["on_{0}".format(method.lower())] = interface

        api_function.interface = interface
        interface.api_function = api_function
        interface.output_format = output
        interface.example = example

        return api_function
    return decorator


for method in HTTP_METHODS:
    globals()[method.lower()] = partial(call, accept=(method, ))



"""hug/run.py

Contains logic to enable execution of hug APIS from the command line
"""
from wsgiref.simple_server import make_server

import json
import falcon
import sys
import importlib
from collections import namedtuple, OrderedDict
from hug import documentation


def documentation_404(module):
    def handle_404(request, response, *kargs, **kwargs):
        base_url = request.url[:-1]
        if request.path and request.path != "/":
            base_url = request.url.split(request.path)[0]
        to_return = OrderedDict()
        to_return['404'] = ("The API call you tried to make was not defined. "
                            "Here's a definition of the API to help you get going :)")
        to_return['documentation'] = documentation.generate(module, base_url)
        response.data = json.dumps(to_return).encode('utf8')
        response.status = falcon.HTTP_NOT_FOUND
    return handle_404


def server(module, sink=documentation_404):
    api = falcon.API()
    for url, method_handlers in module.HUG_API_CALLS.items():
        api.add_route(url, namedtuple('Router', method_handlers.keys())(**method_handlers))
    if sink:
        api.add_sink(sink(module))
    return api


def terminal():
    if len(sys.argv) < 2:
        print("Please specify a hug API file to start the server with")
        sys.exit(1)

    api = server(importlib.machinery.SourceFileLoader(sys.argv[1].split(".")[0], sys.argv[1]).load_module())
    httpd = make_server('', 8000, api)
    print("Serving on port 8000...")
    httpd.serve_forever()


if __name__ == '__main__':
    terminal()



"""hug/types.py

Defines hugs built-in supported types
"""
def number(value):
    """A whole number"""
    return int(value)


def list(value):
    """Multiple Values"""
    return type(value) == list and value or [value]


def comma_separated_list(value):
    """Multiple values, separated by a comma"""
    return value.split(",")


def decimal(value):
    """A decimal number"""
    return float(value)


def text(value):
    """Basic text / string value"""
    return str(value)
